//*******************************************************
//
//	GPS Logger
//
//	MCU:ATmega328p
//		efuse:0xFD	BOD 2.7V
//		hfuse:0xD1	EEPROM消去保護
//		lfuse:0xE2	内蔵CR発振 8MHz
//
//	Compiler:	AVR-GCC		4.3.4
//				AVR-LIBC	1.6.7
//				binutils	2.20
//
//	2010.5.2	KAWAKAMI Yukio
//*******************************************************

#define		VERSION "1.0.7"

#define		LOG_TIME		5		// 秒ごとにログ出力 1～60

#include	<avr/io.h>
#include	<avr/interrupt.h>
#include	<avr/pgmspace.h>
#include	<stdio.h>
#include	<stdlib.h>
#include	<string.h>
#include	<ctype.h>

#include	"GPSLogger.h"

#include	"lcd.h"
#include	"com.h"

#include	"diskio.h"
#include	"ff.h"

extern uchar	CurX, CurY;

volatile uchar	WaitTimer;

uchar		LogBuff[80];

GPS_DATA	Gps;
FATFS		FatFs;
FIL			Log;

//********************************************************************
//	タイマー２比較一致割り込み
//	8MHz/1024/78≒100Hz
//********************************************************************
ISR(TIMER2_COMPA_vect){

	if (WaitTimer)	--WaitTimer;

	disk_timerproc();
}


//********************************************************************
//	メインルーチン
//********************************************************************
int main(void){
	FRESULT	res;
	char	old_sec;

	cli();

	PORTB = (1<<7)|(1<<6)|(1<<MOSI)|(1<<CCS)|(1<<1)|(1<<0);
	DDRB = (1<<SCK)|(1<<MOSI)|(1<<CCS);

	PORTC = 0xFF;
	DDRC = 0;

	PORTD = (1<<LCS)|(1<<SW1)|(1<<PPS)|(1<<TXD)|(1<<RXD);
	DDRD = (1<<LRS)|(1<<LCS)|(1<<LRES)|(1<<POWER);
	
	if (MCUSR & (1<<WDRF)){
		// ウォッチドッグリセット
	}
	MCUSR = 0;

	WDR;
	//WDTCSR = (1<<WDCE)|(1<<WDE);	// WDCEとWDEを1にしてタイマー変更をする
	//WDTCSR = (1<<WDE)|(1<<WDP2)|(1<<WDP1)|(1<<WDP0);	// ウォッチドッグタイマーセット 約２秒

	TCCR2A = (1<<WGM21);					// CTC動作
	TCCR2B = (1<<CS22)|(1<<CS21)|(1<<CS20);	// プリスケーラ 1/1024
	OCR2A = (CPU_CLOCK/1024/100) - 1;		// タイマカウント最大値 100Hz
	TIMSK2 = (1<<OCIE2A);	// 比較Ａ割り込み許可

	SMCR = (1<<SE);			// スリープ許可 アイドルモード

	// TWI, TIMER0, TIMER1 電力カット
	PRR = (1<<PRTWI)|(1<<PRTIM0)|(1<<PRTIM1);
	ACSR = (1<<ACD);			// アナログコンパレータ禁止

	sei();

	wait_timer(20);			// 200ms時間待ち
	
	PORTD |= (1<<POWER);	// POWER Lock

	disk_initialize(0);		// 先にカードを初期化しておく
	
	LCD_init();
	LCD_contrast(40);		// コントラスト
	print_P(PSTR("GPS Logger\nVersion " VERSION "\n yuki-lab.jp"));

	// GM-316 ボーレート初期値 4800bps
	init_com(4800);
	
	while(!(PIND & (1<<SW1))){
		WDR;
	}

	// ボーレート変更コマンド
	GPS_command(PSTR("$PSRF100,1,38400,8,1,0"));
	// 送信完了待ちしてからこちらのボーレート変更
	while(!(UCSR0A & (1<<TXC0)));
	init_com(38400);

	wait_timer(10);	// ちょっと待つ

	GPS_command(PSTR("$PSRF103,03,00,00,01"));	// $GPGSV禁止
	
	clear_LCD();
	CurX = 5;
	CurY = 24;

	// 送信禁止。ひょっとして MCU→GPSモジュールのノイズを低減できるかも
	while(!(UCSR0A & (1<<TXC0)));
	UCSR0B &= ~(1<<TXEN0);

	if ((res = f_mount(0, &FatFs)) == FR_OK){
		DIR		dir;
		char	dirname[8];
		strcpy_P(dirname, PSTR("GPSLOG"));
		res = f_opendir(&dir, dirname);
		if (res == FR_NO_PATH){
			if ((res = f_mkdir(dirname)) != FR_OK){
				print_P(PSTR("mkdir Error"));
				wait_timer(200);
			}
		} else if (res != FR_OK){
			char	buff[32];
			sprintf_P(buff, PSTR("No Card Er.%d"), res);
			print(buff);
			wait_timer(200);
		}
		if (res == FR_OK){
			if (f_chdir(dirname) != FR_OK){
				print_P(PSTR("chdir Error"));
				wait_timer(200);
			}
		}
	} else {
		char	buff[32];
		sprintf_P(buff, PSTR("No Card Er.%d"), res);
		print(buff);
		wait_timer(200);
	}

	old_sec = 0;
	for(;;){

		WDR;
		SLEEP;

		job_serial();
		
		if (Gps.sec != old_sec){
			char	buff[17];
			long	l;

			old_sec = Gps.sec;

			CurX = 5;
			CurY = 0;
			sprintf_P(buff, PSTR("%d/%02d/%02d  %c"),
					  Gps.year, Gps.month, Gps.day, Gps.status);
			print(buff);

			sprintf_P(buff, PSTR("%02d:%02d:%02d"),
					  Gps.hour, Gps.min, Gps.sec);
			print_lerge_num(6, 8, buff);	// 大きな文字で表示

			CurX = 5;
			CurY = 24;
			sprintf_P(buff, PSTR("%c%3d`%02d'%02d.%d\""),
					  Gps.lat_ew, Gps.lat_d, Gps.lat_m,
					  Gps.lat_s/10, Gps.lat_s%10);
			print(buff);

			CurX = 5;
			CurY = 32;
			sprintf_P(buff, PSTR("%c%3d`%02d'%02d.%d\""),
					  Gps.long_ns, Gps.long_d, Gps.long_m,
					  Gps.long_s/10, Gps.long_s%10);
			print(buff);

			CurX = 5;
			CurY = 40;
			sprintf_P(buff, PSTR("   %4d.%dkm/h "),
					  Gps.speed/10, Gps.speed%10);
			if ((Gps.direction > 337)||(Gps.direction < 23)){
				buff[0] = 'N';
			} else if (Gps.direction < 68){
				buff[0] = 'N';
				buff[1] = 'E';
			} else if (Gps.direction < 113){
				buff[0] = 'E';
			} else if (Gps.direction < 158){
				buff[0] = 'S';
				buff[1] = 'E';
			} else if (Gps.direction < 203){
				buff[0] = 'S';
			} else if (Gps.direction < 248){
				buff[0] = 'S';
				buff[1] = 'W';
			} else if (Gps.direction < 293){
				buff[0] = 'W';
			} else {
				buff[0] = 'N';
				buff[1] = 'W';
			}
			print(buff);

			CurX = 5;
			CurY = 48;
			sprintf_P(buff, PSTR("Sat.%2d %5dm "),
					  Gps.satellites, Gps.altitude);
			print(buff);

			// ログファイル作成
			if ((Log.flag == 0)&&
				(FatFs.fs_type)&&
				(Gps.status == 'A')){
				char	fname[13];
				char	c;
				for (c = 'A'; c <= 'Z'; c++){
					sprintf_P(fname, PSTR("%d%02d%02d%02d%c.LOG"),
							  Gps.year%10, Gps.month, Gps.day, Gps.hour, c);
					res = f_open(&Log, fname, FA_READ|FA_OPEN_EXISTING);
					if (res == FR_NO_FILE)	break;
					if (res != FR_OK)	break;
					f_close(&Log);
				}
				if (res == FR_NO_FILE){
					char	str[32];
					res = f_open(&Log, fname, FA_WRITE|FA_CREATE_NEW);
					strcpy_P(str, PSTR("@GPSLogger/ver" VERSION "/wgs-84"));
					write_log(str);
					f_sync(&Log);
				}
			}

			l = ADC_in(VOLT);	// 電池電圧
			l *= (R10_VALUE+R11_VALUE);
			l /= R11_VALUE;
			l *= 110;
			l /= 1023;
			CurX = 5;
			CurY = 56;
			sprintf_P(buff, PSTR("Bat.%ld.%02ldV %2d.%d"),
					  l/100, l%100, Gps.hdop/10, Gps.hdop%10);
			print(buff);

			if (l <= 200){	// 2.00V以下
				power_off();
			}
		}

		if (!(PIND & (1<<SW1))){	// スイッチチェック
			wait_timer(5);
			if (!(PIND & (1<<SW1))){	// スイッチチェック
				power_off();
			}
		}
	}
}

PROGMEM char MonthDays[]
	= { 0, 31, 28, 31, 30, 31, 30, 31, 31, 30, 31, 30, 31 };

//********************************************************************
//	GM-316 受信データ処理
//********************************************************************
void job_serial(void){
	static uchar	com_buff[128];
	static uchar	com_buff_p = 0;

	while(check_com()){
		uchar	data;

		data = read_com();
		if (com_buff_p == 0){
			if (data == '$'){
				com_buff[0] = data;
				com_buff[1] = 0;
				com_buff_p = 1;
			}
		} else if (com_buff_p < (sizeof(com_buff)-1)){
			if (data == LF){
				if (strncmp_P((char *)com_buff, PSTR("$GPGGA"), 6) == 0){
					uchar	buff[12];
					char	*p;
					get_str(com_buff, 7, buff, sizeof(buff));	// 衛星数
					Gps.satellites = atoi((char *)buff);
					get_str(com_buff, 8, buff, sizeof(buff));	// HDOP
					Gps.hdop = atoi((char *)buff) * 10;
					p = strchr((char *)buff, '.');
					if (p != NULL){
						p++;
						Gps.hdop += *p - '0';
					}
					get_str(com_buff, 9, buff, sizeof(buff));	// 高度
					Gps.altitude = atoi((char *)buff);

					if ((atoi((char *)&com_buff[11]) % LOG_TIME) == 0){
						if (com_buff_p < (sizeof(LogBuff) - 1)){
							strcpy((char *)LogBuff, (char *)com_buff);
						}
					}
			
				} else if (strncmp_P((char *)com_buff, PSTR("$GPVTG"), 6) == 0){
					uchar	buff[12];
					char	*p;
					get_str(com_buff, 1, buff, sizeof(buff));	// 移動方向
					Gps.direction = atoi((char *)buff);
					p = strchr((char *)buff, '.');
					if (p != NULL){
						p++;
						if (isdigit(*p) && (*p >= '5')){
							if (++Gps.direction >= 360){
								Gps.direction = 0;
							}
						}
					}
					get_str(com_buff, 7, buff, sizeof(buff));	// スピード
					Gps.speed = atoi((char *)buff) * 10;
					p = strchr((char *)buff, '.');
					if (p != NULL){
						p++;
						if (isdigit(*p)){
							Gps.speed += *p - '0';
						}
					}
					if ((Gps.speed < 0)||(Gps.speed >= 10000)){
						Gps.speed = 0;
					}

					if (LogBuff[0] &&(Gps.status == 'A')&&(Gps.hdop < 100)){
						// ログ記録
						write_log((char *)com_buff);	// $GPVTG
						f_sync(&Log);				// $GPVTGが最後に来るので
					}
					LogBuff[0] = 0;
			
				} else if (strncmp_P((char *)com_buff, PSTR("$GPRMC"), 6) == 0){
					uchar	buff[12];
					char	tm[4];
					long	l;
					char	*p;
					get_str(com_buff, 1, buff, sizeof(buff));	// 時間
					memset(tm, 0, sizeof(tm));
					strncpy(tm, (char *)buff, 2);
					Gps.hour = atoi(tm);
					strncpy(tm, (char *)&buff[2], 2);
					Gps.min = atoi(tm);
					strncpy(tm, (char *)&buff[4], 2);
					Gps.sec = atoi(tm);

					get_str(com_buff, 2, buff, sizeof(buff));	// Status
					Gps.status = buff[0];
					
					get_str(com_buff, 4, buff, 2);	// 北緯・南緯
					if ((buff[0] == 'N')||(buff[0] == 'S')){
						Gps.lat_ew = buff[0];
					
						get_str(com_buff, 3, buff, sizeof(buff));	// 緯度
						l = atol((char *)buff);
						Gps.lat_d = l / 100;
						Gps.lat_m = l % 100;
						p = strchr((char *)buff, '.');
						if (p != NULL){
							p++;
							Gps.lat_s = (atol((char *)p) * 600L) / 10000L;
						} else {
							Gps.lat_s = 0;
						}

						get_str(com_buff, 6, buff, 2);	// 東経・西経
						Gps.long_ns = buff[0];

						get_str(com_buff, 5, buff, sizeof(buff));	// 経度
						l = atol((char *)buff);
						Gps.long_d = l / 100;
						Gps.long_m = l % 100;
						p = strchr((char *)buff, '.');
						if (p != NULL){
							p++;
							Gps.long_s = (atol((char *)p) * 600L) / 10000L;
						} else {
							Gps.long_s = 0;
						}
					}

					get_str(com_buff, 9, buff, sizeof(buff));	// 日付
					strncpy(tm, (char *)buff, 2);
					Gps.day = atoi(tm);
					strncpy(tm, (char *)&buff[2], 2);
					Gps.month = atoi(tm);
					strncpy(tm, (char *)&buff[4], 2);
					Gps.year = atoi(tm) + 2000;

					Gps.hour += 9;	// 日本時間補正
					if (Gps.hour >= 24){
						char	days;
						Gps.hour -= 24;
						Gps.day++;
						days = pgm_read_byte(&MonthDays[(uchar)Gps.month]);
						if ((Gps.month == 2)&&((Gps.year % 4) == 0)){
							days = 29;
						}
						if (Gps.day > days){
							Gps.day = 1;
							Gps.month++;
							if (Gps.month > 12){
								Gps.month = 1;
								Gps.year++;
							}
						}
					}

					if (LogBuff[0] &&(Gps.status == 'A')&&(Gps.hdop < 100)){
						// ログ記録
						write_log((char *)LogBuff);		// $GPGGA
						write_log((char *)com_buff);	// $GPRMC
					}
				} else {
					// ログ記録
					//write_log((char *)com_buff);
				}

				com_buff[0] = 0;
				com_buff_p = 0;
			} else if (data != CR){
				com_buff[com_buff_p++] = data;
				com_buff[com_buff_p] = 0;
			}
		}
	}
}


//********************************************************************
//	電源OFF
//********************************************************************
void power_off(void){

	// ログ Close
	if (Log.flag & FA_WRITE){
		f_close(&Log);
	}

	clear_LCD();
	CurX = 24;
	CurY = 24;
	print_P(PSTR("Power OFF"));

	PORTD &= ~(1<<POWER);	// POWER OFF
	for(;;){
		WDR;
		SLEEP;
	}
}


//********************************************************************
//	ログ記録
//********************************************************************
void write_log(char *str){

	if (Log.flag & FA_WRITE){
		UINT	n;
		char	buff[3];
		f_write(&Log, str, strlen(str), &n);
		strcpy_P((char *)buff, PSTR("\r\n"));
		f_write(&Log, buff, 2, &n);
	}
}


//********************************************************************
//	カンマ区切りの文字列の所定の位置の文字列抽出
//********************************************************************
char get_str(uchar *source, char n, uchar *dest, char max){
	uchar	d;
	char	flag = 0;

	while(n > 0){
		source = (uchar *)strchr((char *)source, ',');
		if (source == NULL)	return 0;
		source++;
		--n;
	}

	--max;
	while((d = *source++)){
		if (d == ',')	break;
		if (max <= 0)	break;
		*dest++ = d;
		flag = 1;
	}
	*dest = 0;
	return flag;
}


//********************************************************************
//	GM-316 コマンド送信
//	コマンド文字列は PSTR("")で指定する
//	*とチェックサムと改行コードは補完する
//********************************************************************
void GPS_command(PGM_P str){
	uchar	sum;
	uchar	d;
	char	buff[6];

	sum = '$';	// 先頭の $ はチェックサムに入れない
	while((d = pgm_read_byte(str++))){
		sum ^= d;
		write_com(d);
	}
	sprintf(buff, "*%02X\r\n", sum);
	write_str((uchar *)buff);
}


//*******************************************************
//	Timerを用いた時間待ち
//*******************************************************
void wait_timer(uchar t){

	WaitTimer = t;
	while(WaitTimer){
		WDR;
		SLEEP;
	}
}


//********************************************************************
//	FatFs 時刻取得
//********************************************************************
DWORD get_fattime(void){
	DWORD	tm;

	tm = ((DWORD)Gps.year - 1980) << 25;
	tm |= ((DWORD)Gps.month) << 21;
	tm |= ((DWORD)Gps.day) << 16;
	tm |= ((DWORD)Gps.hour) << 11;
	tm |= ((DWORD)Gps.min) << 5;
	tm |= Gps.sec / 2;
	return tm;
}


//********************************************************************
//	SPIデータ出力
//********************************************************************
void SPI_out(uchar data){

	SPDR = data;
	while(!(SPSR & (1<<SPIF)));
}


//********************************************************************
//	SPIデータ入力
//********************************************************************
uchar SPI_in(void){

	SPDR = 0xFF;
	while(!(SPSR & (1<<SPIF)));
	return SPDR;
}


//********************************************************************
//	A/D変換
//********************************************************************
int ADC_in(char ch){
	int	data;
	
	// A/D変換準備 1.1V基準
	ADMUX = (1<<REFS0)|(1<<REFS1)|ch;
	// A/D変換開始 プリスケーラ 8MHz/64 = 125kHz
	ADCSRA = (1<<ADEN)|(1<<ADPS1)|(1<<ADPS2);
	ADCSRA |= (1<<ADSC);	// 変換開始
	while(!(ADCSRA & (1<<ADIF)));	// 変換完了待ち
	ADCSRA |= (1<<ADIF);	// ADIFクリア
	data = ADCL;
	data |= ((int)ADCH) << 8;
	return data;
}

